module GapFilling
using LinearAlgebra, Statistics, HypothesisTests
using Arpack:svds

export gap_filling
export cal_reconstruct_rms, ini_missing

function arpack_svd(A; k=1)
    res = svds(A, nsv=k)[1]
    return res.U, res.S, res.V
end

function vector_to_Hankel(x, L::Int)
    m = length(x)
    res = Matrix{eltype(x)}(undef, m - L + 1, L)
    for i in 1:L
        res[:, i] = x[i:(m-L+i)]
    end
    return res
end


function Hankel_to_vector(X)
    M, L = size(X)
    N = L + M - 1
    res = zeros(N)
    count = zeros(N)
    for j in 1:L, i in 1:M
        k = i + j - 1
        res[k] += X[i, j]
        count[k] += 1
    end
    return res ./ count
end

function svd_to_matrix(U, S, V)
    R = size(U, 1)
    C = size(V, 1)
    m = size(S, 1)
    if R < C
        U1 = similar(U)
        for i in 1:m
            U1[:, i] = U[:, i] .* S[i]
        end
        return U1 * V'
    else
        V1 = similar(V)
        for i in 1:m
            V1[:, i] = V[:, i] .* S[i]
        end
        return U * V1'
    end
end

function svd_to_matrix(U, S, V, k::Int)
    return svd_to_matrix(U[:, 1:k], S[1:k], V[:, 1:k])
end


function extract_signal(y, L; k=2)
    A = vector_to_Hankel(y, L)
    U, S, V = arpack_svd(A, k=k)
    X_hat = svd_to_matrix(U, S, V)
    x_hat1 = Hankel_to_vector(X_hat)
    return x_hat1
end


function symmetric_test(x)
    m = median(x)
    x1 = m .- x[x.<m]
    x2 = x[x.>m] .- m
    return pvalue(MannWhitneyUTest(x1, x2))
end

function pvalue_is_noise(y; w=30)
    A = vector_to_Hankel(y, w)
    B = A' * A
    res_eigen, tmp = eigen(B)
    pvalue = symmetric_test(res_eigen)
    return pvalue
end


function rms(x)
    sqrt(mean(x .^ 2))
end

function rms(x, idx)
    rms(x[idx])
end

function rms(y1, y2, idx_miss)
    rms(y1[idx_miss] - y2[idx_miss])
end

function impute_by_one_SSA(x, L; k=2)
    A = vector_to_Hankel(x, L)
    U, S, V = arpack_svd(A, k=k)
    X_low_rank = svd_to_matrix(U, S, V, k)
    x_hat = Hankel_to_vector(X_low_rank)
    return x_hat
end

function rowmean(x)
    mean(x, dims=2)[:, 1]
end


function ini_missing(y, idx_miss)
    win_len = length(y) ÷ 2
    m0 = mean(y[.!idx_miss])
    y_old = copy(y) .- m0
    y_new = similar(y_old)
    i = 1
    while true
        y_imp = impute_by_one_SSA(y_old, win_len, k=4)
        y_new = y_old .* (1 .- idx_miss) .+ y_imp .* idx_miss
        rmse = rms(y_old - y_new, idx_miss)
        y_old = copy(y_new)
        i = i + 1
        i > 30 && break
        rmse < 0.01 && break
    end
    return y_new .+ m0
end


function reconstruct_seq(y, win_len, idx_missing, k)
    i = 1
    m0 = mean(y)
    x_old = copy(y) .- m0
    x_new = similar(x_old)
    x_new0 = similar(x_old)
    while true
        x_new0 = extract_signal(x_old, win_len, k=k)
        x_new = x_old .* (1 .- idx_missing) .+ x_new0 .* idx_missing
        mse = rms(x_old[idx_missing] - x_new[idx_missing])
        mse < 0.01 && break
        #i > 10  && break
        i = i + 1
        x_old = x_new
    end
    return x_new0 .+ m0
end

function impute_missing_by_avg(y, idx_miss; k=10)
    win_len_vec = 2 .^ (4:1:floor(Int, log2(length(y) / 2)))
    res = zeros(length(y), length(win_len_vec))
    m0 = mean(y[.!idx_miss])
    y1 = y .- m0
    y1_ini = ini_missing(y1, idx_miss)
    for i in 1:length(win_len_vec)
        y_imputed = reconstruct_seq(y1_ini, win_len_vec[i], idx_miss, k)
        res[:, i] = y_imputed .+ m0
    end
    return rowmean(res)
end


function gap_filling(y)

    idx_miss = ismissing.(y)
    m0 = mean(skipmissing(y))
    y = y .- m0
    y[idx_miss] .= 0.0

    i = 0
    y_resid = convert.(Float64, y)
    y_imputed = fill(0.0, length(y))
    pval = 0.01
    while pval < 0.05
        i = i + 1
        y_imputed = y_imputed + impute_missing_by_avg(y_resid, idx_miss; k=2)
        y_resid = y - y_imputed
        pval = pvalue_is_noise(y_resid[.!idx_miss], w=100)
        res_rms = rms(y, y_imputed, idx_miss)
#        println("i= $i, rms = $res_rms and p-value =", round(pval, digits=3))
        i > 10 && break
    end
    return y_imputed .+ m0
end


function cal_reconstruct_rms(y1, y0, win_len_vec)

    rms_vec = fill(0.0, length(win_len_vec))
    i = 1
    for w in win_len_vec
        y_reconst = extract_signal(y1, w, k=4)
        res_rms = rms(y_reconst - y0)
        rms_vec[i] = res_rms
        i = i + 1
    end
    return rms_vec
end

function cal_reconstruct_rms(y1, y0, idx_miss, win_len_vec)
    y1_ini = copy(y1)
    rms_vec = fill(0.0, length(win_len_vec))
    i = 1
    for w in win_len_vec
        y_imputed = reconstruct_seq(y1_ini, w, idx_miss, 4)
        res_rms = rms(y0, y_imputed, idx_miss)
        rms_vec[i] = res_rms
        i = i + 1
    end
    return rms_vec
end


end

